-- This software is copyright Kong Inc. and its licensors.
-- Use of the software is subject to the agreement between your organization
-- and Kong Inc. If there is no such agreement, use is governed by and
-- subject to the terms of the Kong Master Software License Agreement found
-- at https://konghq.com/enterprisesoftwarelicense/.
-- [ END OF LICENSE 0867164ffc95e54f04670b5169c09574bdbd9bba ]

local parse = require"kong.gql.query.build_ast.parse"
local nodes = require"kong.gql.query.nodes"
local QueryAst = require"kong.gql.query.ast"


-- TODO: Refactor list map patterns below by substituting for this subroutine.
local function map_lst(lst, lambda)
    if not lst or not lambda then return nil end
    local r = {}
    for _, i in ipairs(lst) do
        table.insert(r, lambda(i))
    end
    return r
end


local function _arguments(arguments)
    if not arguments then return nil end
    local r = {}
    for _, arg in ipairs(arguments) do
        local value
        if arg.value.kind == 'float' or arg.value.kind == 'int' then
            value = tonumber(arg.value.value)
        elseif arg.value.kind == 'boolean' then
            if arg.value.value == 'true' then value = true else value = false end
        else value = arg.value.value end

        r[arg.name.value] = value
    end
    return r
end


local function _directives(dir_arr)
    if not dir_arr then return nil end
    local r = map_lst(
            dir_arr,
            function (dir)
                local arguments = _arguments(dir.arguments)
                return nodes.types.Directive(dir.name.value, arguments)
            end
    )
    return r
end


local function _alias(alias)
    if not alias then return nil end
    return alias.name.value
end


-- Possibilities are field, fragment spread, or inline fragment
local function selection_node(s)
    if s.kind == "field" then
        local name = s.name.value
        local arguments = _arguments(s.arguments)
        local selection_nodes = {}
        if s.selectionSet and s.selectionSet.selections then
            for _, selection in ipairs(s.selectionSet.selections) do
                table.insert(selection_nodes, selection_node(selection))
            end
        end
        local directives = _directives(s.directives)
        local alias = _alias(s.alias)
        return nodes.FieldNode(name, arguments, selection_nodes, directives, alias)
    elseif s.kind == "inlineFragment" then
        local type_condition = s.typeCondition.name.value
        local selection_nodes = map_lst(s.selectionSet.selections, selection_node)
        local directives = _directives(s.directives)
        return nodes.InlineFragmentNode(type_condition, selection_nodes, directives)
    elseif s.kind == "fragmentSpread" then
        local directives = _directives(s.directives)
        return nodes.FragmentSpread(s.name.value, directives)
    end
end


local function _selection_set(set)
    if not set then return nil end
    local r = {}
    for _, s in ipairs(set.selections) do
        table.insert(r, selection_node(s))
    end
    return r
end


local function _var_defs(var_defs)
    if not var_defs then return nil end
    local d_list = {}
    for _, vd in ipairs(var_defs) do
        -- TODO: Parse default value in variables
        local vd_name = vd.variable.name.value
        local vd_type
        if vd.type.kind == 'namedType' then
            vd_type = vd.type.name.value
        else
            if vd.type.type.kind == 'namedType' then
                vd_type = vd.type.type.name.value
            elseif vd.type.type.kind == 'nonNullType' then
                vd_type = vd.type.type.type.name.value
            end
        end
        local vd_n = nodes.types.VarDef(vd_name, vd_type, nil)
        table.insert(d_list, vd_n)
    end
    return d_list
end


-- Entry point into conversion of parse tree to
-- native AST.
-- @param st: a BjornBytes parse tree
local function deserialize_parse_tree(st)
    if st.kind == "document" then
        local def_list = map_lst(st.definitions, deserialize_parse_tree)
        return nodes.DocumentNode(def_list)
    elseif st.kind == "operation" then
        local _type = st.operation
        local selection_nodes = _selection_set(st.selectionSet)
        local name = st.name and st.name.value
        local var_defs = _var_defs(st.variableDefinitions)
        local directives = _directives(st.directives)
        return nodes.OperationNode(_type, selection_nodes, name, var_defs, directives)
    elseif st.kind == "fragmentDefinition" then
        local selection_nodes = _selection_set(st.selectionSet)
        local directives = _directives(st.directives)
        local type_condition = st.typeCondition.name.value
        local name = st.name.value
        return nodes.FragmentDefNode(name, type_condition, selection_nodes, directives)
    end
end


return function (string, schema)
    local parse_tree = parse(string)
    local document_tree = deserialize_parse_tree(parse_tree)
    local ast = QueryAst.new(document_tree)

    return ast
end
