-- This software is copyright Kong Inc. and its licensors.
-- Use of the software is subject to the agreement between your organization
-- and Kong Inc. If there is no such agreement, use is governed by and
-- subject to the terms of the Kong Master Software License Agreement found
-- at https://konghq.com/enterprisesoftwarelicense/.
-- [ END OF LICENSE 0867164ffc95e54f04670b5169c09574bdbd9bba ]

local ecdsa          = require "resty.nettle.ecdsa"
local eddsa          = require "resty.nettle.eddsa"
local hmac           = require "resty.nettle.hmac"
local sha2           = require "resty.nettle.sha2"
local rsa            = require "resty.nettle.rsa"
local codec          = require "kong.openid-connect.codec"
local nyi            = require "kong.openid-connect.nyi"
local openssl_pkey   = require "resty.openssl.pkey"


local base64url = codec.base64url
local json      = codec.json
local sub       = string.sub
local fmt       = string.format
local type      = type


local function errmsg(msg, err)
  if err then
    return fmt("%s (%s)", msg, err)
  end

  return msg
end


local S256 = sha2.sha256
local S384 = sha2.sha384
local S512 = sha2.sha512


local ES_CURVES = {
  ["P-192"] = true,
  ["P-224"] = true,
  ["P-256"] = true,
  ["P-384"] = true,
  ["P-521"] = true,
}


local ES_SIZES = {
  ["P-192"] = 24,
  ["P-224"] = 28,
  ["P-256"] = 32,
  ["P-384"] = 48,
  ["P-521"] = 66,
}


local ED_CURVES = {
  ["Ed25519"] = true,
  ["Ed448"] = true,
}


local HASH = {
  S256 = S256,
  S384 = S384,
  S512 = S512,
  HS256 = S256,
  HS384 = S384,
  HS512 = S512,
  RS256 = S256,
  RS384 = S384,
  RS512 = S512,
  ES256 = S256,
  ES384 = S384,
  ES512 = S512,
  PS256 = S256,
  PS384 = S384,
  PS512 = S512,
}


local HS = {}


function HS.sign(alg, jwk, inp)
  local k = jwk.k
  if not k then
    return nil, "hs shared secret was not specified"
  end

  local err
  k, err = base64url.decode(k)
  if not k then
    return nil, errmsg("hs key value could not be base64 url decoded", err)
  end

  local dgt
  -- todo, inp == nil fails with traceback
  dgt, err = alg(k, inp)
  if not dgt then
    return nil, errmsg("unable to sign hs digest", err)
  end

  dgt, err = base64url.encode(dgt)
  if not dgt then
    return nil, errmsg("unable to base64 url encode hs signature", err)
  end

  return dgt
end


function HS.verify(alg, jwk, inp, sig)
  local k = jwk.k
  if not k then
    return nil, "hs shared secret was not specified"
  end

  if not sig then
    return nil, "hs signature was not specified"
  end

  local err
  sig, err = base64url.decode(sig)
  if not sig then
    return nil, errmsg("hs signature could not be base64 url decoded", err)
  end

  k, err = base64url.decode(k)
  if not k then
    return nil, errmsg("hs key could not be base64 url decoded", err)
  end

  local mac
  mac, err = alg(k, inp)
  if not mac then
    return nil, errmsg("unable to count hs digest", err)
  end

  if mac == sig then
    return true
  end

  return nil, "hs signature verification failed"
end


local RS = {}


function RS.sign(alg, jwk, inp)
  local d = jwk.d
  if not d then
    return nil, "rs secret exponent was not specified"
  end

  local p = jwk.p
  if not p then
    return nil, "rs factor p was not specified"
  end

  local q = jwk.q
  if not q then
    return nil, "rs factor q was not specified"
  end

  local err
  local n = jwk.n
  if n then
    n, err = base64url.decode(n)
    if not n then
      return nil, errmsg("rs modulus could not be base64 url decoded", err)
    end
  end

  local e = jwk.e
  if e then
    e, err = base64url.decode(e)
    if not e then
      return nil, errmsg("rs exponent could not be base64 url decoded", err)
    end
  end

  local pub
  if e and n then
    pub, err = rsa.key.public.new(n, e)
    if not pub then
      return nil, errmsg("rs public key could not be prepared", err)
    end
  end

  d, err = base64url.decode(d)
  if not d then
    return nil, errmsg("rs secret exponent could not be base64 url decoded", err)
  end

  p, err = base64url.decode(p)
  if not p then
    return nil, errmsg("rs factor p could not be base64 url decoded", err)
  end

  q, err = base64url.decode(q)
  if not q then
    return nil, errmsg("rs factor q could not be base64 url decoded", err)
  end

  local dp = jwk.dp
  if dp then
    dp, err = base64url.decode(dp)
    if not dp then
      return nil, errmsg("rs factor dp could not be base64 url decoded", err)
    end
  end

  local dq = jwk.dq
  if dq then
    dq, err = base64url.decode(dq)
    if not dq then
      return nil, errmsg("rs factor dq could not be base64 url decoded", err)
    end
  end

  local qi = jwk.qi
  if qi then
    qi, err = base64url.decode(qi)
    if not qi then
      return nil, errmsg("rs factor qi could not be base64 url decoded", err)
    end
  end

  local pri
  pri, err = rsa.key.private.new(d, p, q, dp, dq, qi)
  if not pri then
    return nil, errmsg("rs private key could not be prepared", err)
  end

  local r
  r, err = rsa.new(pub, pri)
  if not r then
    return nil, errmsg("unable to prepare rs algorithm", err)
  end

  local dgt
  dgt, err = alg(inp)
  if not dgt then
    return nil, errmsg("unable to count rs digest", err)
  end

  dgt, err = r:sign_digest(dgt)
  if not dgt then
    return nil, errmsg("unable to sign rs digest", err)
  end

  dgt, err = base64url.encode(dgt)
  if not dgt then
    return nil, errmsg("unable to base64 url encode rs signature", err)
  end

  return dgt
end


function RS.verify(alg, jwk, inp, sig)
  local n = jwk.n
  if not n then
    return nil, "rs modulus was not specified"
  end

  local e = jwk.e
  if not e then
    return nil, "rs exponent was not specified"
  end

  local err

  n, err = base64url.decode(n)
  if not n then
    return nil, errmsg("rs modulus could not be base64 url decoded", err)
  end
  e, err = base64url.decode(e)
  if not n then
    return nil, errmsg("rs exponent could not be base64 url decoded", err)
  end

  local p
  p, err = rsa.key.public.new(n, e)
  if not p then
    return nil, errmsg("rs public key could not be prepared", err)
  end

  local s
  s, err = base64url.decode(sig)
  if not s then
    return nil, errmsg("rs signature could not be base64 decoded", err)
  end

  local d
  d, err = alg(inp)
  if not d then
    return nil, errmsg("unable to count rs digest", err)
  end

  local r
  r, err = rsa.new(p)
  if not r then
    return nil, errmsg("unable to prepare rs algorithm", err)
  end

  return r:verify_digest(d, s)
end


function RS.encrypt()
  return nyi()
end


function RS.decrypt()
  return nyi()
end


local PS = {}


function PS.sign(alg, jwk, inp)
  local n = jwk.n
  if not n then
    return nil, "ps modulus was not specified"
  end

  local e = jwk.e
  if not e then
    return nil, "ps exponent was not specified"
  end

  local d = jwk.d
  if not d then
    return nil, "ps secret exponent was not specified"
  end

  local p = jwk.p
  if not p then
    return nil, "ps factor p was not specified"
  end

  local q = jwk.q
  if not q then
    return nil, "ps factor q was not specified"
  end

  local err
  n, err = base64url.decode(n)
  if not n then
    return nil, errmsg("ps modulus could not be base64 url decoded", err)
  end

  e, err = base64url.decode(e)
  if not n then
    return nil, errmsg("ps exponent could not be base64 url decoded", err)
  end

  local pub
  pub, err = rsa.key.public.new(n, e)
  if not pub then
    return nil, errmsg("ps public key could not be prepared", err)
  end

  d, err = base64url.decode(d)
  if not d then
    return nil, errmsg("ps secret exponent could not be base64 url decoded", err)
  end

  p, err = base64url.decode(p)
  if not p then
    return nil, errmsg("ps factor p could not be base64 url decoded", err)
  end

  q, err = base64url.decode(q)
  if not q then
    return nil, errmsg("ps factor q could not be base64 url decoded", err)
  end

  local dp = jwk.dp
  if dp then
    dp, err = base64url.decode(dp)
    if not dp then
      return nil, errmsg("ps factor dp could not be base64 url decoded", err)
    end
  end

  local dq = jwk.dq
  if dq then
    dq, err = base64url.decode(dq)
    if not dq then
      return nil, errmsg("ps factor dq could not be base64 url decoded",  err)
    end
  end

  local qi = jwk.qi
  if qi then
    qi, err = base64url.decode(qi)
    if not qi then
      return nil, errmsg("ps factor qi could not be base64 url decoded", err)
    end
  end

  local pri
  pri, err = rsa.key.private.new(d, p, q, dp, dq, qi)
  if not pri then
    return nil, errmsg("ps private key could not be prepared", err)
  end

  local r
  r, err = rsa.new(pub, pri)
  if not r then
    return nil, errmsg("unable to prepare ps algorithm", err)
  end

  local dgt
  dgt, err = alg(inp)
  if not dgt then
    return nil, errmsg("unable to count ps digest", err)
  end

  dgt, err = r:pss_sign_digest(dgt)
  if not dgt then
    return nil, errmsg("unable to sign ps digest", err)
  end

  dgt, err = base64url.encode(dgt)
  if not dgt then
    return nil, errmsg("unable to base64 url encode ps signature", err)
  end

  return dgt
end


function PS.verify(alg, jwk, inp, sig)
  local n = jwk.n
  if not n then
    return nil, "ps modulus was not specified"
  end

  local e = jwk.e
  if not e then
    return nil, "ps exponent was not specified"
  end

  local err
  n, err = base64url.decode(n)
  if not n then
    return nil, errmsg("ps modulus could not be base64 url decoded", err)
  end

  e, err = base64url.decode(e)
  if not n then
    return nil, errmsg("ps exponent could not be base64 url decoded", err)
  end

  local p
  p, err = rsa.key.public.new(n, e)
  if not p then
    return nil, errmsg("ps public key could not be prepared", err)
  end

  local s
  s, err = base64url.decode(sig)
  if not s then
    return nil, errmsg("ps signature could not be base64 url decoded", err)
  end

  local d
  d, err = alg(inp)
  if not d then
    return nil, errmsg("unable to count ps digest", err)
  end

  local r
  r, err = rsa.new(p)
  if not r then
    return nil, err
  end

  return r:pss_verify_digest(d, s)
end


function PS.encrypt()
  return nyi()
end


function PS.decrypt()
  return nyi()
end


local ES = {}


function ES.sign(alg, jwk, inp)
  local err

  local crv = jwk.crv
  if not crv then
    return nil, "es curve was not specified"
  end

  if not ES_CURVES[crv] then
    return nil, errmsg("invalid or unsupported es curve", crv)
  end

  local d = jwk.d
  if not d then
    return nil, "es ecc private key was not specified"
  end

  d, err = base64url.decode(d)
  if not d then
    return nil, errmsg("es ecc private key could not be base64 url decoded", err)
  end

  local sz
  sz, err = ecdsa.scalar.new(crv, d)
  if not sz then
    return nil, errmsg("es scalar could not be prepared", err)
  end

  local dgt
  dgt, err = alg(inp)
  if not dgt then
    return nil, errmsg("unable to count es digest", err)
  end

  local ecd
  ecd, err = ecdsa.new(nil, sz)
  if not ecd then
    return nil, errmsg("es algorithm could not be prepared", err)
  end

  dgt, err = ecd:sign(dgt, ES_SIZES[crv])
  if not dgt then
    return nil, errmsg("unable to ecdsa sign es digest", err)
  end

  dgt, err = base64url.encode(dgt.r .. dgt.s)
  if not dgt then
    return nil, errmsg("unable to base64 url encode es signature", err)
  end

  return dgt
end


function ES.verify(alg, jwk, inp, sig)
  local x = jwk.x
  if not x then
    return nil, "ecdsa x coordinate was not specified"
  end

  local y = jwk.y
  if not y then
    return nil, "ecdsa y coordinate was not specified"
  end

  local crv = jwk.crv
  if not crv then
    return nil, "ecdsa curve was not specified"
  end

  if not ES_CURVES[crv] then
    return nil, errmsg("invalid or unsupported ecdsa curve", crv)
  end

  local exp, err = base64url.decode(sig)
  if not exp then
    return nil, errmsg("ecdsa signature could not be base64 url decoded", err)
  end

  local len = #exp
  if len % 2 ~= 0 then
    return nil, "invalid ecdsa signature"
  end

  local mid = len / 2
  local r = sub(exp, 1, mid)
  local s = sub(exp, mid + 1)

  x, err = base64url.decode(x)
  if not x then
    return nil, errmsg("es x coordinate could not be base64 url decoded", err)
  end

  y, err = base64url.decode(y)
  if not y then
    return nil, errmsg("es y coordinate could not be base64 url decoded", err)
  end

  local pxy
  pxy, err = ecdsa.point.new(crv, x, y)
  if not pxy then
    return nil, errmsg("es point could not be prepared", err)
  end

  local srs
  srs, err = ecdsa.signature.new(r, s)
  if not srs then
    return nil, errmsg("es signature could not be prepared", err)
  end

  local dgt
  dgt, err = alg(inp)
  if not dgt then
    return nil, errmsg("unable to count es digest", err)
  end

  local ecd
  ecd, err = ecdsa.new(pxy)
  if not ecd then
    return nil, errmsg("es algorithm could not be prepared", err)
  end

  return ecd:verify(dgt, srs)
end


function ES.encrypt()
  return nyi()
end


function ES.decrypt()
  return nyi()
end


local EdDSA = {}


function EdDSA.sign(jwk, inp)
  local crv = jwk.crv
  if not crv then
    return nil, "ed curve was not specified"
  end

  if not ED_CURVES[crv] then
    return nil, errmsg("invalid or unsupported eddsa curve", crv)
  end

  local d = jwk.d
  if not d then
    return nil, "ed private key was not specified"
  end

  local err
  d, err = base64url.decode(d)
  if not d then
    return nil, errmsg("ed private key could not be base64 url decoded", err)
  end

  local x = jwk.x
  if x then
    x, err = base64url.decode(x)
    if not x then
      return nil, errmsg("ed public key could not be base64 url decoded", err)
    end
  end

  local dgt
  dgt, err = eddsa.sign(x, d, inp)
  if not dgt then
    return nil, errmsg("unable to eddsa sign", err)
  end

  dgt, err = base64url.encode(dgt)
  if not dgt then
    return nil, errmsg("unable to base64 url encode ed signature", err)
  end

  return dgt
end


function EdDSA.verify(jwk, inp, sig)
  local x = jwk.x
  if not x then
    return nil, "eddsa x coordinate was not specified"
  end

  local crv = jwk.crv
  if not crv then
    return nil, "eddsa curve was not specified"
  end

  if not ED_CURVES[crv] then
    return nil, errmsg("invalid or unsupported eddsa curve", crv)
  end

  local exp, err = base64url.decode(sig)
  if not exp then
    return nil, errmsg("eddsa signature could not be base64 url decoded", err)
  end

  x, err = base64url.decode(x)
  if not x then
    return nil, errmsg("ed x coordinate could not be base64 url decoded", err)
  end

  return eddsa.verify(x, inp, exp)
end


local HS256 = {}


function HS256.sign(...)
  return HS.sign(hmac.sha256, ...)
end


function HS256.verify(...)
  return HS.verify(hmac.sha256, ...)
end


local HS384 = {}


function HS384.sign(...)
  return HS.sign(hmac.sha384, ...)
end


function HS384.verify(...)
  return HS.verify(hmac.sha384, ...)
end


local HS512 = {}


function HS512.sign(...)
  return HS.sign(hmac.sha512, ...)
end


function HS512.verify(...)
  return HS.verify(hmac.sha512, ...)
end


local RS256 = {}


function RS256.sign(...)
  return RS.sign(S256, ...)
end


function RS256.verify(...)
  return RS.verify(S256, ...)
end


local RS384 = {}


function RS384.sign(jwk, inp)
  jwk.kty = "RSA"
  local json_jwk, err = json.encode(jwk)
  if not json_jwk then
    return nil, err
  end
  local pkey, pkeyerr = openssl_pkey.new(json_jwk, {format = "JWK"})
  if not pkey then
    return nil, pkeyerr
  end
  local sig = assert(pkey:sign(inp, "SHA384"))
  return base64url.encode(sig)
end

function RS384.verify(jwk, inp, sig)
  jwk.kty = "RSA"
  local dc_sig, err = base64url.decode(sig)
  if not dc_sig then
    return nil, errmsg("signature could not be base64 url decoded", err)
  end
  local json_jwk, ec_err = json.encode(jwk)
  if not json_jwk then
    return nil, ec_err
  end
  local pkey, pkeyerr = openssl_pkey.new(json_jwk, {format = "JWK"})
  if not pkey then
    return nil, pkeyerr
  end
  return pkey:verify(dc_sig, inp, "sha384")
end


local RS512 = {}


function RS512.sign(...)
  return RS.sign(S512, ...)
end


function RS512.verify(...)
  return RS.verify(S512, ...)
end


local ES256 = {}


function ES256.sign(...)
  return ES.sign(S256, ...)
end


function ES256.verify(...)
  return ES.verify(S256, ...)
end


local ES384 = {}


function ES384.sign(...)
  return ES.sign(S384, ...)
end


function ES384.verify(...)
  return ES.verify(S384, ...)
end


local ES512 = {}


function ES512.sign(...)
  return ES.sign(S512, ...)
end


function ES512.verify(...)
  return ES.verify(S512, ...)
end


local PS256 = {}


function PS256.sign(...)
  return PS.sign(S256, ...)
end


function PS256.verify(...)
  return PS.verify(S256, ...)
end


local PS384 = {}


function PS384.sign(...)
  return PS.sign(S384, ...)
end


function PS384.verify(...)
  return PS.verify(S384, ...)
end


local PS512 = {}


function PS512.sign(...)
  return PS.sign(S512, ...)
end


function PS512.verify(...)
  return PS.verify(S512, ...)
end


local HMAC = {
  HS256 = HS256,
  HS384 = HS384,
  HS512 = HS512,
}


local SIGN = {
  HS256 = HS256.sign,
  HS384 = HS384.sign,
  HS512 = HS512.sign,
  RS256 = RS256.sign,
  RS384 = RS384.sign,
  RS512 = RS512.sign,
  ES256 = ES256.sign,
  ES384 = ES384.sign,
  ES512 = ES512.sign,
  PS256 = PS256.sign,
  PS384 = PS384.sign,
  PS512 = PS512.sign,
  EdDSA = EdDSA.sign,
}


local VERIFY = {
  HS256 = HS256.verify,
  HS384 = HS384.verify,
  HS512 = HS512.verify,
  RS256 = RS256.verify,
  RS384 = RS384.verify,
  RS512 = RS512.verify,
  ES256 = ES256.verify,
  ES384 = ES384.verify,
  ES512 = ES512.verify,
  PS256 = PS256.verify,
  PS384 = PS384.verify,
  PS512 = PS512.verify,
  EdDSA = EdDSA.verify,
}


local jwa = {
  CURVES = ES_CURVES,
  HASH   = HASH,
  HMAC   = HMAC,
  SIGN   = SIGN,
  VERIFY = VERIFY,
  S256   = S256,
  S384   = S384,
  S512   = S512,
  HS256  = HS256,
  HS384  = HS384,
  HS512  = HS512,
  RS256  = RS256,
  RS384  = RS384,
  RS512  = RS512,
  ES256  = ES256,
  ES384  = ES384,
  ES512  = ES512,
  PS256  = PS256,
  PS384  = PS384,
  PS512  = PS512,
  EdDSA  = EdDSA,
}


function jwa.hash(alg, inp)
  local hash = HASH[alg]
  if not hash then
    return nil, "unsupported jwa hashing algorithm"
  end
  return hash(inp)
end


function jwa.verify(alg, jwk, ...)
  if not alg then
    return nil, "jwa signature verification algorithm was not specified"
  end

  if jwk.alg then
    if jwk.alg ~= alg then
      return nil, "algorithm mismatch"
    end
  end

  local verify = VERIFY[alg]

  if not verify then
    return nil, "unsupported jwa signature verification algorithm was specified"
  end

  if type(jwk) ~= "table" then
    return nil, "invalid jwk was supplied for jwa signature verification"
  end

  return verify(jwk, ...)
end


function jwa.sign(alg, jwk, ...)
  if not alg then
    return nil, "jwa signing algorithm was not specified"
  end

  local sign = SIGN[alg]

  if not sign then
    return nil, "unsupported jwa signing algorithm was specified"
  end

  if type(jwk) ~= "table" then
    return nil, "invalid jwk was supplied for jwa signing"
  end

  return sign(jwk, ...)
end


return jwa
